import 'dart:math';

import 'package:germanet_dart/src/germanet.dart';
import 'package:germanet_dart/src/lexunit.dart';
import 'package:germanet_dart/src/synset.dart';

class Filterconfig {
  String search_string;
  bool ignore_case;
  bool regex;
  int levenshtein_distance;
  late List word_categories;
  late List word_classes;
  late List orth_variants;
  Filterconfig(this.search_string,{ this.ignore_case = false, this.regex=false, this.levenshtein_distance = 0}){
    word_categories = WordCategory.values;
    word_classes = WordClass.values;
    orth_variants = OrthFormVariant.values;
  }

  ///filter synsets to a specific search_string, categroy, class, orth_variant (see init)
  Set<Synset> filter_synsets(Germanet germanet) {
    Set<Synset> result = {};
    Set<Lexunit> lexunits;
    if (regex) {
      lexunits = _get_lexunits_by_regex(germanet);
    }else{
      if(levenshtein_distance > 0 && !regex){
        lexunits = _filter_lexunits_levenshtein(germanet);
      }else{
        lexunits = germanet.getLexunitsByOrthform(search_string, ignorecase: ignore_case).toSet();
        lexunits = _filter_lexunits_orthform(lexunits, orth_variants, search_string, ignore_case);
      }
    }
    List<Synset> synsets = lexunits.map((e) => e.synset).toList();
    for (Synset synset in synsets) {
      if(word_classes.contains(synset.word_class) && word_categories.contains(synset.category)){
        result.add(synset);
      }
    }
    return result;
  }
  
  Set<Lexunit> _get_lexunits_by_regex(Germanet germanet) {

    Set<Lexunit> result = {};
    List l_ids;
    if(ignore_case){
      RegExp pattern = RegExp(search_string);
      l_ids = germanet.lowercasedform2lexid.keys.where((element) => pattern.hasMatch(element)).toList().map((e) => germanet.lowercasedform2lexid[e]).toList();
    }else{
      RegExp pattern = RegExp(search_string);
      l_ids = germanet.orthform2lexid.keys.where((element) => pattern.hasMatch(element)).toList().map((e) => germanet.orthform2lexid[e]).toList();
    }
    for (String id in l_ids) {
      result.add(germanet.lexunits[id]);
    }
    return result;

  }
  
  Set<Lexunit> _filter_lexunits_levenshtein(Germanet germanet) {
    Set<Lexunit> filtered_lexunits = {};
    for (WordCategory cat in word_categories) {
      List<Lexunit> units = germanet.getLexunitsByWordcategory(cat);
      for (Lexunit unit in units) {
        if(word_classes.contains(unit.synset.word_class)){
          for (OrthFormVariant orthvar in orth_variants) {
            var form = unit.get_orthform_variant(orthvar);
            if(form != null && form != ""){
              if(ignore_case){
                form = form.toLowerCase();
                search_string = search_string.toLowerCase();
              }
              int d = distance(form, search_string);
              if(d <= levenshtein_distance){
                filtered_lexunits.add(unit);
              }
            }
          }
        }
      }
    }
    return filtered_lexunits;
  }
  
  int distance(String s, String t, {bool caseSensitive: true}) {
  if (!caseSensitive) {
    s = s.toLowerCase();
    t = t.toLowerCase();
  }
  if (s == t) {
    return 0;
  }
  if (s.isEmpty) {
    return t.length;
  }
  if (t.isEmpty) {
    return s.length;
  }

  List<int> v0 = List<int>.filled(t.length + 1, 0);
  List<int> v1 = List<int>.filled(t.length + 1, 0);

  for (int i = 0; i < t.length + 1; i < i++) {
    v0[i] = i;
  }

  for (int i = 0; i < s.length; i++) {
    v1[0] = i + 1;

    for (int j = 0; j < t.length; j++) {
      int cost = (s[i] == t[j]) ? 0 : 1;
      v1[j + 1] = min(v1[j] + 1, min(v0[j + 1] + 1, v0[j] + cost));
    }

    for (int j = 0; j < t.length + 1; j++) {
      v0[j] = v1[j];
    }
  }

  return v1[t.length];
}

  Set<Lexunit> _filter_lexunits_orthform(Set<Lexunit> lexunits, List orth_variants, String search_string, bool ignore_case) {

    Set<Lexunit> filtered_units = {};
    for (Lexunit unit in lexunits) {
      for (OrthFormVariant orth_variant in orth_variants) {
        String form = unit.get_orthform_variant(orth_variant);
        if(form == search_string){
          filtered_units.add(unit);
        }
        if(ignore_case && form != null && form != ""){
          if(form.toLowerCase() == search_string.toLowerCase()){
            filtered_units.add(unit);
          }
        }
      }
    }
    return filtered_units;

  }


}