import 'dart:io';

import 'package:germanet_dart/germanet_dart.dart';
import 'package:germanet_dart/src/iliRecord.dart';
import 'package:germanet_dart/src/pseudo_file.dart';
import 'package:germanet_dart/src/synsetLoader.dart';
import 'package:germanet_dart/src/utils.dart';
import 'package:germanet_dart/src/wiktionaryparaphrase.dart';
import 'package:xml/xml.dart';

import 'compoundinfo.dart';
import 'frames.dart';

class Germanet {

  String sNAME = 'name';
  String sFROMNODE = 'from';
  String sTONODE = 'to';
  String sDIRECTION = 'dir';
  String sINVERSE = 'inv';
  String sLEXREL = 'lex_rel';
  String sCONREL = 'con_rel';
  String sLEXID = 'lexUnitId';
  String sID = 'wiktionaryId';
  String sSENSEID = 'wiktionarySenseId';
  String sSENSE = 'wiktionarySense';
  String sEDITED = 'edited';
  String sEWNREL = 'ewnRelation';
  String sPWNWORD = 'pwnWord';
  String sPWN20ID = 'pwn20Id';
  String sPWN30ID = 'pwn30Id';
  String sSOURCE = 'source';
  String sPWN20PARAPHRASE = 'pwn20paraphrase';

  String datadir;
  bool addIlirecords;
  bool addWictionary;
  late Map lexunits;
  late Map synsets;
  late Map orthform2lexid;
  late Map mainOrtform2lexid;
  late Map lowercasedform2lexid;
  late Map wordcat2lexid;
  late Map wordclass2lexid;
  late Set compounds;
  late Map frames2lexunits;
  late List wiktionaryEntries;
  late List iliRecords;
  late Frames frames;

  /// The GermaNet object is initialized with the directory where the GermaNet data is stored. The data is loaded
  /// when GermaNet is initialized.
  /// String datadir: The path to the directory where the GermaNet data is stored
  /// bool addIlirecords: a boolean, denotes whether the iliRecords should also be loaded into the GermaNet
  /// bool addWictionary: a boolean, denotes whether the wiktionary files should also be loaded into the GermaNet
  /// bool loadDataDirectly: whether to load the data directly
  Germanet(this.datadir, {this.addIlirecords = true, this.addWictionary = true, bool loadDataDirectly = true}){
    
    lexunits = {};
    synsets = {};
    orthform2lexid = {};
    mainOrtform2lexid = {};
    lowercasedform2lexid = {};
    wordcat2lexid = {};
    wordclass2lexid = {};
    compounds = {};
    frames2lexunits = {};
    wiktionaryEntries = [];
    iliRecords = [];
    frames = Frames(frames2lexunits);

    if(loadDataDirectly){
      loadData();
    }
  }

  ///loading the GermanNet files (at datadir).
  ///This can take up to 10 seconds
  ///a Future<void> instance will be returned
  Future<void> loadData({GermanetMessage? germanetMessage, SecurityCryptInterpreter? securityCryptInterpreter, bool usePseudoFiles = false, List<PseudoFile>? pseudoFiles}) async{
    if(germanetMessage != null){
      germanetMessage.sendCurrentStatus(0.0, "Loading Source Files");
    }
    List files = usePseudoFiles ? pseudoFiles! : await Directory(datadir).list().toList();  
    List wikifiles = files.where((element) => element.path.contains("wiktionary") && element.path.endsWith(".xml")).toList();
    List lexentries = files.where((element) => (element.path.contains("nomen") || element.path.contains("verben") || element.path.contains("adj"))  && element.path.endsWith(".xml")).toList();
    List ilifiles = files.where((element) => element.path.contains("interLingua")  && element.path.endsWith(".xml")).toList();

    if(germanetMessage != null){
      germanetMessage.sendCurrentStatus(0.00441, "Loading GermaNet Lexunits");
    }
    for (var i = 0; i < lexentries.length; i++) {
      var f = lexentries[i];
      XmlDocument document = parseXml(f, securityCryptInterpreter); //tree
      await _loadLexunits(document); // system return fixed /01
    }

    if(germanetMessage != null){
      germanetMessage.sendCurrentStatus(0.78435, "Loading GermaNet Word Relations");
    }

    XmlDocument tree = parseXml(files.firstWhere((element) => element.path.endsWith("gn_relations.xml")), securityCryptInterpreter);
    _loadRelations(tree);
    if(addWictionary){
      if(germanetMessage != null){
      germanetMessage.sendCurrentStatus(0.99960, "Loading Wiktionary Data");
    }
      for (var i = 0; i < wikifiles.length; i++) {
        XmlDocument tree = parseXml(wikifiles[i], securityCryptInterpreter);
        await _loadWiktionary(tree);
      }
    }
    if(addIlirecords){
      if(germanetMessage != null){
      germanetMessage.sendCurrentStatus(0.99987, "Loading Ili Records");
    }
      for (var i = 0; i < ilifiles.length; i++) {
        XmlDocument tree = parseXml(ilifiles[i], securityCryptInterpreter);
        await _loadIli(tree);
      }
    }
  }
  Future<void> _loadLexunits(XmlDocument tree) async{
  XmlElement root = tree.rootElement;
  for (XmlElement child in root.childElements) {
    List<XmlAttribute> attrs = child.attributes;
    String synId = attrs.firstWhere((element) => element.localName == 'id').value;
    dynamic stCat = attrs.firstWhere((element) => element.localName == 'category').value;
    WordCategory category = WordCategory.values.firstWhere((element) => element.name == stCat);
    stCat = attrs.firstWhere((element) => element.localName == 'class').value;
    WordClass wordClass = WordClass.values.firstWhere((element) => element.name == stCat);

    Synset synset = Synset(synId, category, wordClass);
    for (XmlElement subChild in child.childElements) {
      // ignore: unrelated_type_equality_checks
      if(subChild.name.qualified == 'lexUnit'){
        Lexunit lexunit = _createLexunit(subChild.attributes, subChild, synset);
        lexunits[lexunit.id] = lexunit;
        if(wordcat2lexid.containsKey(category.name)){
          wordcat2lexid[category.name].add(lexunit.id);
        }else{
          wordcat2lexid[category.name] = [lexunit.id];
        }
        if(wordclass2lexid.containsKey(wordClass.name)){
          wordclass2lexid[wordClass.name].add(lexunit.id);
        }else{
          wordclass2lexid[wordClass.name] = [lexunit.id];
        }
        synset.add_lexunit(lexunit);
      }
    }
    for (Lexunit unit in synset.lexunits) {
        for (Lexunit lexunit in synset.lexunits) {
          if(unit != lexunit){
            if(unit.relations.containsKey(LexRel.has_synonym.name)){
               unit.relations[LexRel.has_synonym.name].add(lexunit);
               unit.synset.relations[LexRel.has_synonym.name].add(lexunit.synset);
            }else{
               unit.relations[LexRel.has_synonym.name] = [lexunit];
               unit.synset.relations[LexRel.has_synonym.name] = [lexunit.synset];
            }
          }
        }
    }

    synsets[synset.id] = synset;
    //print(synsets.length);
  }
  
}
Lexunit _createLexunit(List<XmlAttribute> attributes, XmlNode lexRoot, Synset synset){
  String lexId = attributes.firstWhere((element) => element.localName == "id").value;
  int lexSense = int.parse(attributes.firstWhere((element) => element.localName == "sense").value);
  String lexSource = attributes.firstWhere((element) => element.localName == "source").value;
  bool lexNamedEntity = attributes.firstWhere((element) => element.localName == "namedEntity").value == "true";
  bool lexArtificial = attributes.firstWhere((element) => element.localName == "artificial").value == "true";
  bool lexStyle = attributes.firstWhere((element) => element.localName == "styleMarking").value == "true";
  Lexunit lexunit = Lexunit(lexId, synset, lexSense, lexSource, lexNamedEntity, lexStyle, lexArtificial);
  for (XmlElement child in lexRoot.childElements) {
    String tag = child.name.qualified;
    String childValue = child.text;
    if(tag == 'compound'){
      if(child.attributes.isEmpty){
          continue;
      }
        CompoundInfo compoundInfo = create_compound_info(child);
        lexunit.compound_info = compoundInfo;
        compounds.add(lexunit);
    }else{
      if(tag.contains("rth")){
        _addOrthForms(lexunit, childValue, tag);
      }else{
        if(tag == "frame"){
          lexunit.frames.add(childValue);
          for (var f in lexunit.frames) {
            if(frames2lexunits.containsKey(f)){
              frames2lexunits[f].add(lexunit);
            }else{
              frames2lexunits[f] = [lexunit];
            }
          }
        }else{
          if(tag == "example"){
            if(child.attributes.isEmpty){
          continue;
      }
            var example = child.attributes.first.text;
            lexunit.examples.add(example);
            if(child.attributes.length == 2){
              var exframe = child.attributes[1].text;
              lexunit.frames2examples[exframe].add(example);
            }
          }
        }
      }
    }
  }
  return lexunit;
}
void _addOrthForms(Lexunit lexunit, String childValue, String tag){

  if(orthform2lexid.containsKey(childValue)){
    orthform2lexid[childValue].add(lexunit.id);
  }else{
    orthform2lexid[childValue] = [lexunit.id];

  }
  if(lowercasedform2lexid.containsKey(childValue.toLowerCase())){
    lowercasedform2lexid[childValue.toLowerCase()].add(lexunit.id);
  }else{
    lowercasedform2lexid[childValue.toLowerCase()] = [lexunit.id];
  }

  if(tag == "orthForm"){
    lexunit.orthform = childValue;
    if(mainOrtform2lexid.containsKey(childValue)){
      mainOrtform2lexid[childValue].add(lexunit.id);
    }else{
      mainOrtform2lexid[childValue] = [lexunit.id];
    }
  }else{
    if(tag == "orthVar"){
      lexunit.orthvar = childValue;
    }else{
      if(tag == "oldOrthForm"){
        lexunit.old_orthform = childValue;
      }else{
        if(tag == "oldOrthVar"){
          lexunit.old_othvar = childValue;
        }
      }
    }
  }
}

Future<void> _loadRelations(XmlDocument tree) async{

  XmlNode root = tree.root;
  for (XmlElement child in root.childElements.first.childElements) {
      String tag = child.name.qualified;
      if(tag == sLEXREL){
        var relation = child.attributes.firstWhere((element) => element.name.qualified == sNAME).value;
        var fromNode = child.attributes.firstWhere((element) => element.name.qualified == sFROMNODE).value;
        var toNode = child.attributes.firstWhere((element) => element.name.qualified == sTONODE).value;
        //var direction = child.attributes.firstWhere((element) => element.name.qualified == sDIRECTION).value;
        Lexunit lexunit2 = lexunits[toNode];
        Lexunit lexunit1 = lexunits[fromNode];
        var lx0 = LexRel.values.firstWhere((element) => element.value == relation);
        if(lexunit1.relations.containsKey(lx0)){
          lexunit1.relations[lx0].add(lexunit2);
        }else{
          lexunit1.relations[lx0] = [lexunit2];
        }
        if(lexunit2.incoming_relations.containsKey(lx0)){
          lexunit2.incoming_relations[lx0].add(lexunit1);
        }else{
          lexunit2.incoming_relations[lx0] = [lexunit1];
        }
        if(child.attributes.any((p0) => p0.name.qualified == sINVERSE)){
          var inverseRelation = child.attributes.firstWhere((element) => element.name.qualified == sINVERSE).value;
          lx0 = LexRel.values.firstWhere((element) => element.value == inverseRelation);
          if(lexunit2.relations.containsKey(lx0)){
            lexunit2.relations[lx0].add(lexunit1);      
          }else{
            lexunit2.relations[lx0] = [lexunit1];
          }
        }
      }else{
        if(tag == sCONREL){
          if(child.attributes.isEmpty){
            continue; // ext.0
          }
          var relation = child.attributes.firstWhere((element) => element.name.qualified == sNAME).value;
          var fromNode = child.attributes.firstWhere((element) => element.name.qualified == sFROMNODE).value;
          var toNode = child.attributes.firstWhere((element) => element.name.qualified == sTONODE).value;
          //var direction = child.attributes.firstWhere((element) => element.name.qualified == sDIRECTION).value;
          Synset synset1 = synsets[fromNode];
          Synset synset2 = synsets[toNode];
          var lx0 = ConRel.values.firstWhere((element) => element.name == relation);
          if(synset1.relations.containsKey(lx0)){
            synset1.relations[lx0].add(synset2);
          }else{
            synset1.relations[lx0] = [synset2];
          }
          if(synset2.incoming_relations.containsKey(lx0)){
            synset2.incoming_relations[lx0].add(synset1);
          }else{
            synset2.incoming_relations[lx0] = [synset1];
          }
          if(child.attributes.any((p0) => p0.name.qualified == sINVERSE)){
            var inverseRelation = child.attributes.firstWhere((element) => element.name.qualified == sINVERSE).value;
            if(synset2.relations.containsKey(ConRel.values.firstWhere((element) => element.name == inverseRelation))){
              synset2.relations[ConRel.values.firstWhere((element) => element.name == inverseRelation)].add(synset1);
            }else{
              synset2.relations[ConRel.values.firstWhere((element) => element.name == inverseRelation)] = [synset1];
            }
          }
        }
      }
  }

}


Future<void> _loadWiktionary(XmlDocument tree) async {

  XmlNode root = tree.root;
  for (XmlNode child in root.childElements) {
      List<XmlAttribute> attrs = child.attributes;
      if(attrs.isEmpty){
        continue;
      }
      var lexId = attrs.firstWhere((element) => element.name.qualified == sLEXID).value;
      var wiktionaryId = attrs.firstWhere((element) => element.name.qualified == sID).value;
      var wiktionarySenseId = int.parse(attrs.firstWhere((element) => element.name.qualified == sSENSEID).value);
      var wiktionarySense = attrs.firstWhere((element) => element.name.qualified == sSENSE).value;
      var edited = attrs.firstWhere((element) => element.name.qualified == sEDITED).value == "true";
      WiktionaryParaphrase wiktionaryParaphrase = WiktionaryParaphrase(lexId, wiktionaryId, wiktionarySenseId, wiktionarySense, edited);
      Lexunit lexunit = getLexunitById(wiktionaryParaphrase.lexunit_id);
      lexunit.wiktionary_paraphrases.add(wiktionaryParaphrase);
      wiktionaryEntries.add(wiktionaryParaphrase);
  }

}

Future<void> _loadIli(XmlDocument tree) async{
  XmlNode root = tree.root;
  for (XmlElement child in root.childElements) {
    List synonyms = [];
    for (XmlElement subchild in child.childElements) {
      for (XmlElement subsubchild in subchild.childElements) {
        synonyms.add(subsubchild.text);
      }
    }
    if(child.attributes.isEmpty){
      //print("continue");
      continue;
    }
    List<XmlAttribute> attributes = child.attributes;
    String lexId = attributes.firstWhere((element) => element.qualifiedName == sLEXID).value;
    String pwnrelation = (attributes.firstWhere((element) => element.qualifiedName == sEWNREL).value);
    String pwnword = (attributes.firstWhere((element) => element.qualifiedName == sPWNWORD).value);
    String pwn20Id = attributes.firstWhere((element) => element.qualifiedName == sPWN20ID).value;
    String pwn30Id = attributes.firstWhere((element) => element.qualifiedName == sPWN30ID).value;
    String source = attributes.firstWhere((element) => element.qualifiedName == sSOURCE).value;
    String pwn20paraphrase = attributes.firstWhere((element) => element.qualifiedName == sPWN20PARAPHRASE).value;
    
  
    IliRecord ili = IliRecord(lexId, pwnrelation, pwnword, pwn20Id, pwn30Id, source, synonyms, pwn20paraphrase: pwn20paraphrase);
    Lexunit lexunit = lexunits[ili.lexunit_id];
    lexunit.ili_records.add(ili);
    iliRecords.add(ili);

  }
}
/// This method returns a list of synsets that match the given input search string
List<Synset> getSynsetsByOrthform(String form,  {bool ignorecase = false}){

  List lexunitIds = [];
  if(ignorecase){
    form = form.toLowerCase();
    lexunitIds = lowercasedform2lexid[form];
  }else{
    lexunitIds = orthform2lexid[form];
  }
  List<Synset> lxs = [];
  for (var id in lexunitIds) {
    lxs.add(lexunits[id].synset);
  }
  return lxs;
}

bool exists(String form, {bool ignorecase = false}){
  if(ignorecase){
    return lowercasedform2lexid.containsKey(form.toLowerCase());
  }else{
    return orthform2lexid.containsKey(form);
  }
}

/// Returns a list of synsets that belong to the specified word category
List<Synset> getSynsetsByWordcategory(WordCategory category){

  List<Synset> lxs = [];
  for (var id in wordcat2lexid[category.name]) {
    lxs.add(lexunits[id].synset);
  }
  return lxs;

}

/// Returns a list of synsets that belong to the specified word class
List<Synset> getSynsetsByWordclass(WordClass wordClass){
  List<Synset> lxs = [];
  for (Lexunit id in wordclass2lexid[wordClass.name]) {
    lxs.add(id.synset);
  }
  return lxs;
}

/// Returns a Synset by a specified identifier (if that exists, otherwise raises an Error)
Synset getSynsetById(String id){
  if(synsets.containsKey(id)){

    return synsets[id];

  }else{

    throw Exception("the given Synset id is not in GermaNet");

  }
}

/// Returns a lexical unit by a specified identifier (if that exists, otherwise raises an Error)
Lexunit getLexunitById(String id){

  if(lexunits.containsKey(id)){

    return lexunits[id];

  }else{
    throw Exception("the given lexical unit id is not in GermaNet");
  }

}

/// This method returns a list of lexical units that match the given input search string
List<Lexunit> getLexunitsByOrthform(String form, {bool ignorecase = false}){

  if(ignorecase){
    form = form.toLowerCase();
    var lexunitIds = lowercasedform2lexid[form];
    List<Lexunit> lxs = [];
    for (var id in lexunitIds) {
      lxs.add(lexunits[id]);
    }
    return lxs;
  }
  var lexunitIds = orthform2lexid[form];
  List<Lexunit> lxs = [];
  for (var id in lexunitIds) {
    lxs.add(lexunits[id]);
  }
  return lxs;

}

/// Returns a list of lexical units that belong to the specified word class
List<Lexunit> getLexunitsByWordclass(WordClass classA){

  var lexunitIds = wordclass2lexid[classA.name];
  List<Lexunit> lxs = [];
  for (var id in lexunitIds) {
    lxs.add(lexunits[id]);
  }
  return lxs;
}

/// Returns a list of lexical units that belong to the specified word category
List<Lexunit> getLexunitsByWordcategory(WordCategory category){

  var lexunitIds = wordcat2lexid[category.name];
  List<Lexunit> lxs = [];
  for (var id in lexunitIds) {
    lxs.add(lexunits[id]);
  }
  return lxs;

}

/// Returns a list of Synsets that match a specified frame
/// frame: a frame that describes the argument structure of a verb (e.g. 'NN.AN' specifies that a verb can take a subject and accusative object as arguments.)
List<Synset> getSynsetsByFrame(String frame){

  if(frames2lexunits.containsKey(frames)){

    var synsetIds = frames2lexunits[frame];
    List<Synset> lxs = [];
    for (var id in synsetIds) {
      lxs.add(synsets[id]);
    }
    return lxs;

  }else{
    throw Exception("the specified frame is not in GermaNet");
  }

}

Synset root(){

  return getSynsetById('s51001');

}


}