import 'dart:io';
import 'dart:math';

import 'package:germanet_dart/germanet_dart.dart';
import 'package:germanet_dart/src/filterconfig.dart';
import 'package:germanet_dart/src/functools.dart';
import 'package:germanet_dart/src/semrel_measures.dart';
import 'package:germanet_dart/src/synset.dart';

class ICBasedSimilarity {
  Germanet germanet;
  WordCategory category;
  String path;
  String separator;
  late Map<dynamic, dynamic> synset2simplefreq;
  late Map<dynamic, dynamic> synset2cumfreq;
  late int root_freq;
  late var jcnmaxdist;
  late Map _synset2ic;
  late Synset _most_informative_synset;
  late Map _normalization_dic;

  /// The IC-based measures are computed based on relative frequencies of words in a large corpus. Synset frequencies
  /// are computed by adding up the frequencies of all words that belong to a Synset. These measures can not be computed between
  /// synsets with different word categories
  ICBasedSimilarity(this.germanet, this.category, this.path, {this.separator = "\t"}){

    synset2simplefreq = _init_simple_freq_dic(category);
    _create_simple_freq_dic(category, path, separator);
    synset2cumfreq = synset2simplefreq;
    _compute_cum_freq_dic();

    root_freq = synset2simplefreq[germanet.root().id];
    jcnmaxdist = (2 * - log(1 / root_freq) / log(10));

    List l0 = _init_ic_map();
    _synset2ic = l0[0];
    _most_informative_synset = l0[1];

    List<Synset> synset_pair = [_most_informative_synset, germanet.root()];
    _normalization_dic = _init_min_max_normalization_values(synset_pair);

  }

  Map<String, dynamic> _init_simple_freq_dic(WordCategory category){

    List<Synset> all_synsets_wordcat = germanet.getSynsetsByWordcategory(category);
    Set all_synsets_wordcat2 = all_synsets_wordcat.map((e) => e.id).toSet(); 
    all_synsets_wordcat2.add(germanet.root().id);
    Map<String, dynamic> synset2cumfreq = directZipMap(all_synsets_wordcat2.toList(), List.filled(all_synsets_wordcat2.length, 1)).cast();
    return synset2cumfreq;
  }

  void _create_simple_freq_dic(WordCategory wordCategory, String path, String seperator){

    if(File(path).existsSync()){
      try {
        for (String line in File(path).readAsLinesSync()) {
          List parts = line.trim().split(seperator);
          int freq = int.parse(parts[1]);
          String word = parts[0];
          Filterconfig config = Filterconfig(word);
          config.word_categories = [wordCategory];
          Set<Synset> synsets = config.filter_synsets(germanet);
          for (Synset syn in synsets) {
            synset2simplefreq[syn.id] += freq;
          }
        }
      } catch (e) {
        print(e);
        print("Could not open/read file: "+ path);
        exit(-1);
      }
    }else{
      print("file "+ path + " does not exits");
    }

  }
  
  void _compute_cum_freq_dic() {
    for (var synset_id in synset2simplefreq.keys) {
      int simple_freq = synset2simplefreq[synset_id];
      _cumulate(germanet.getSynsetById(synset_id), simple_freq);
    }
  }
  
  void _cumulate(Synset synsetById, int simple_freq) {
    if(synsetById.is_root()){
      return;
    }
    List<Synset> hypernyms = synsetById.direct_hypernyms();
    if(hypernyms.isNotEmpty){
      for (var hyper in hypernyms) {
        synset2cumfreq[hyper.id] += simple_freq;
        _cumulate(hyper, simple_freq);
      }
    }

  }
  
  List _init_ic_map() {
    Map synset2ic = {};
    double max_ic = 0;
    Synset most_informative_synset = germanet.root();
    for (String synset_id in synset2cumfreq.keys) {
      Synset synset = germanet.getSynsetById(synset_id);
      double ic = _get_information_content(synset);
      synset2ic[synset_id] = ic;
      if(ic > max_ic){
        max_ic = ic;
        most_informative_synset = synset;
      }
    }
    return [synset2ic, most_informative_synset];
  }
  
  double _get_information_content(Synset synset) {
    assert(synset2cumfreq.containsKey(synset.id));
    var synset_freq = synset2cumfreq[synset.id];
    return - (log(synset_freq / root_freq) / log(10));
  }
  
  Map _init_min_max_normalization_values(List<Synset> synsets) => {
      SemRelMeasure.Resnik: [resnik(synsets[0], synsets[1]), resnik(synsets[0], synsets[0])],
      SemRelMeasure.Lin: [lin(synsets[0], synsets[1]), lin(synsets[0], synsets[0])],
      SemRelMeasure.JiangAndConrath: [jiang_and_conrath(synsets[0], synsets[1]), jiang_and_conrath(synsets[0], synsets[0])]
  };
  
  /// Two concepts are more related the more information they share. The shared information of two
  /// concepts can be quantified by the information content of two concepts' lowest common subsumer.
  /// When several LCS are available the highest IC is returned.
  double resnik(Synset synset1, Synset synset2, {bool normalize = false, double normalized_max = 1.0}) {
    double resnik = 0;
    if(synset1 == synset2){
      resnik = _get_information_content(synset1);
    }else{
      var lcs_list = synset1.lowest_common_subsumer(synset2);
      for (var lcs in lcs_list) {
        double ic = _get_information_content(lcs);
        if(ic > resnik){
          resnik = ic;
        }
      }
    }
    if(normalize){
      this.normalize(resnik, normalized_max, SemRelMeasure.Resnik);
    }
    return resnik;
  }

  /// The Jiang and Conraths measure includes knowledge about the individual information contents of each synset.
  /// The smaller the difference of the information content of the two synsets, the more related they are.
  double jiang_and_conrath(Synset synset1, Synset synset2, {bool normalize = false, double normalized_max = 1.0}){
    double distance = _get_information_content(synset1) + _get_information_content(synset2) - 2 * resnik(synset1, synset2);
    double jcn = jcnmaxdist - distance;
    if(normalize){
      jcn = this.normalize(jcn, normalized_max, SemRelMeasure.JiangAndConrath);
    }
    return jcn;
  }

  /// The lin measure takes the individual information contents of each synset and the information content of the
  /// LCS into account. The LCS with the highest information content is used for the computation.
  double lin(Synset synset1, Synset synset2, {bool normalize = false, double normalized_max = 1.0}){
    double ic_lcs = resnik(synset1, synset2);
    double lin = (2 * ic_lcs) / (_get_information_content(synset1) + _get_information_content(synset2));
    if(normalize){
      lin = this.normalize(lin, normalized_max, SemRelMeasure.Lin);
    }
    return lin;
  }

  /// Normalizes a raw value of semantic relatedness to a value between a lower bound and the given upper bound.
  double normalize(double raw_value, double normalized_max, SemRelMeasure semrel_measure){
    List l0 = _normalization_dic[semrel_measure];
    int lower_bound = l0[0];
    int upper_bound = l0[1];
    return dp(((raw_value - lower_bound) / (upper_bound - lower_bound)) * normalized_max, 5);
  }






}

