import 'dart:math';

import 'package:germanet_dart/germanet_dart.dart';
import 'package:germanet_dart/src/longest_shortest_path.dart';
import 'package:germanet_dart/src/semrel_measures.dart';
import 'package:germanet_dart/src/synset.dart';

import 'functools.dart';

class PathBasedRelatedness {
  Germanet germanet;
  WordCategory category;
  int max_len;
  int max_depth;
  List? synset_pair = [];
  late Map normalization_dic;

  /// These measures use the GermaNet Graph to compute the shortest Paths between two concepts. These concepts have to
  /// have the same word category. The path lengths are normalized in different ways (depending on the measure). The path
  /// lengths are computed taking only the hypernymy / hyponymy relations into account
  PathBasedRelatedness(this.germanet, this.category, {this.max_len = -1, this.max_depth = -1, this.synset_pair = null}){
    if(max_len == -1 || max_depth == -1 || synset_pair == null){
      List l0 = get_longest_possible_shortest_distance(germanet, category);
      max_len = l0[0];
      max_depth = l0[1];
      synset_pair = l0[2];
    }
    normalization_dic = _init_min_max_normalization_values(synset_pair!);
  }

  Map _init_min_max_normalization_values(List synsets) => {
      SemRelMeasure.SimplePath: [simple_path(synsets[0], synsets[1]), simple_path(synsets[0], synsets[0])],
      SemRelMeasure.WuAndPalmer: [wu_and_palmer(synsets[0], synsets[1]), wu_and_palmer(synsets[0], synsets[0])],
      SemRelMeasure.LeacockAndChodorow: [leacock_chodorow(synsets[0], synsets[1]), leacock_chodorow(synsets[0], synsets[0])]
  };

  
  /// This methods computes the semantic relatedness by taking the path length into account, normalizing by taking
  /// the depth of the LCS. If there are several possible LCS, the one with the largest depth is taken into account.
  double wu_and_palmer(Synset synset1, Synset synset2, {bool normalize = false, double normalized_max = 1.0}){

    assert(synset1.category == synset2.category);

    Synset root_node = germanet.root();
    var lcs_nodes = synset1.lowest_common_subsumer(synset2);
    int depth = 0;
    for (Synset n in lcs_nodes) {
      int current_depth = n.shortest_path_distance(root_node);
      if(current_depth > depth){
        depth = current_depth;
      }
    }
    int path_len = synset2.shortest_path_distance(synset1);
    double wup = (2 * depth) / (path_len + 2 * depth);
    if(normalize){
      wup = this.normalize(wup, normalized_max, SemRelMeasure.WuAndPalmer);
    }
    return dp(wup, 5);
  }


  /// This method implements the leackock and chodorow relatedness measure. For the path distance and depth,
  /// node count is used.
  double leacock_chodorow(Synset synset1, Synset synset2, {bool normalize = false, double normalized_max = 1.0}){

    assert(synset1.category == synset2.category);

    int pathlen = synset1.shortest_path_distance(synset2) + 1;
    double lch_sim = - (log(pathlen / (2 * (max_depth + 1)))) / log(10);
    if(normalize){
      lch_sim = this.normalize(lch_sim, normalized_max, SemRelMeasure.LeacockAndChodorow);
    }
    return dp(lch_sim, 5);
  } 


  /// This measure computes the pathlength and normalizes it by the longest possible shortest path between any two
  /// nodes of the corresponding word category.
  double simple_path(Synset synset1, Synset synset2, {bool normalize = false, double normalized_max = 1.0}){
    assert(synset1.category == synset2.category);
    int pathlen = synset1.shortest_path_distance(synset2);
    double path = (max_len - pathlen) / max_len;
    if(normalize){

      path = this.normalize(path, normalized_max, SemRelMeasure.SimplePath);

    }
    return dp(path, 5);
  }

  /// Normalizes a raw value of semantic relatedness to a value between a lower bound and the given upper bound.
  double normalize(double raw_value, double normalized_max, SemRelMeasure semrel_measure){
    List l0 = normalization_dic[semrel_measure];
    int lower_bound = l0[0];
    int upper_bound = l0[1];
    return dp(((raw_value - lower_bound) / (upper_bound - lower_bound)) * normalized_max, 5);
  }
}